;-- TIME_STAMP   2018-12-26 08:10:50   v 0.1

#Region    ;************ Includes ************
#include-once
#Include <WinAPIShPath.au3>
#EndRegion ;************ Includes ************

;===============================================================================
; Function Name: _BuildInZIP($sPath, $s_a_ToZip, $fUnZip=True, $iIndexBase=1, $fOverwrite=True)
; Description: Erstellt aus einer Datei od. einem Array von Dateien eine ZIP-Datei
; mit Hilfe der Windows eigenen ZIP-Funktion
; Parameter(s): $sPath      : Pfad der zu erstellenden ZIP-Datei // Bestehenden ZIP-Datei
;               $s_a_ToZip  : String der Einzeldatei oder Array der zu zippenden Dateien
;                             // Pfad des Zielordners, in den entpackt werden soll
;               $fUnZip     : Entpacken=True, Packen=False (Standard)
;               $iIndexBase : Index Basis (erstes Element mit Pfad) des übergebenen Arrays (0 od.1)
;               $fOverwrite : Falls $sPath existiert ==> Überschreiben (Standard=True)
; Return Value(s): Erfolg 1
; Fehler set @error 1 - Überschreiben aus, aber Zieldatei existiert
; 2 - Zieldatei endet nicht auf ".zip"
; 3 - zu komprimierende Datei existiert nicht
; @extended: Anzahl der fehlenden Dateien
; Author(s): BugFix (bugfix@autoit.de)
; Modified : Bitnugger
; Link : https://autoit.de/index.php?thread/19163-zip-packen-entpacken-mit-windows-eigener-funktion/
;===============================================================================
Func _BuildInZIP($sPath, $s_a_ToZip, $fUnZip = False, $iIndexBase = 1, $fOverwrite = True)
	; Error monitoring. This will trap all COM errors while alive.
    ; This particular object is declared as local, meaning after the function returns it will not exist.
    Local $oErrorHandler = ObjEvent("AutoIt.Error", "_ErrFunc")
	If Not IsObj($oErrorHandler) Then Exit 0x3F

;~ 	Local $fh, $sNul = '', $sZIPHeader, $objShell = ObjCreate("Shell.Application"), $objZipOrdner, $err = 0
	Local $fh, $sZIPHeader, $objShell = ObjCreate("Shell.Application"), $objZipOrdner, $err = 0
	; FileGetLongName("filename", 1) funktioniert nur, wenn "filename" existiert!
	$sPath = _WinAPI_PathSearchAndQualify($sPath)
	$s_a_ToZip = _WinAPI_PathSearchAndQualify($s_a_ToZip)
	ConsoleWrite("@@ Debug line" & @TAB & @ScriptLineNumber & "   var: $sPath      --> " & $sPath & @CRLF)
	ConsoleWrite("@@ Debug line" & @TAB & @ScriptLineNumber & "   var: $s_a_ToZip  --> " & $s_a_ToZip & @CRLF)
	If $fUnZip Then
		Local $temp, $objZip
		If Not FileExists($s_a_ToZip) Then DirCreate($s_a_ToZip)
		$temp = $objShell.Namespace($s_a_ToZip)
;~ 		ConsoleWrite("@@ Debug line" & @TAB & @ScriptLineNumber & "   var: $temp   --> " & $temp & ' VarGetType($temp) = ' & VarGetType($temp) & @CRLF)
		$objZip = $objShell.Namespace($sPath)
;~ 		ConsoleWrite("@@ Debug line" & @TAB & @ScriptLineNumber & "   var: $objZip --> " & $objZip & ' VarGetType($objZip) = ' & VarGetType($objZip) & @CRLF)
		$temp.CopyHere($objZip.Items)
		Return 1
	EndIf
	If (Not $fOverwrite) And FileExists($sPath) Then Return SetError(1, 0, 0)
	If StringRight($sPath, 4) <> '.zip' Then Return SetError(2, 0, 0)
;~ 	For $i = 1 To 18
;~ 		$sNul &= Chr(0)
;~ 	Next
;~ 	$sZIPHeader = Chr(80) & Chr(75) & Chr(5) & Chr(6) & $sNul
	$sZIPHeader = Binary('0x504B0506000000000000000000000000000000000000')

	$fh = FileOpen($sPath, 26) ; $FO_OVERWRITE (2) + $FO_CREATEPATH (8) + $FO_BINARY (16)
	FileWrite($fh, $sZIPHeader)
	FileClose($fh)
	$objZipOrdner = $objShell.Namespace($sPath)
	If Not IsArray($s_a_ToZip) Then
		If Not FileExists($s_a_ToZip) Then Return SetError(3, 1, 0)
		$objZipOrdner.CopyHere($s_a_ToZip)
		Sleep(1000)
		Return 1
	EndIf
	For $i = $iIndexBase To UBound($s_a_ToZip) - 1
		If Not FileExists($s_a_ToZip[$i]) Then
			$err += 1
			ContinueLoop
		EndIf
		$objZipOrdner.CopyHere($s_a_ToZip[$i])
		Sleep(1000)
	Next
	If $err Then Return SetError(3, $err, 0)
	Return 1
EndFunc   ;==>_BuildInZIP

; User's COM error function. Will be called if COM error occurs
Func _ErrFunc($oError)
    ; Do anything here.
    ConsoleWrite(@ScriptName & " (" & $oError.scriptline & ") : ==> COM Error intercepted !" & @CRLF & _
            @TAB & "err.number is: " & @TAB & @TAB & "0x" & Hex($oError.number) & @CRLF & _
            @TAB & "err.windescription:" & @TAB & $oError.windescription & @CRLF & _
            @TAB & "err.description is: " & @TAB & $oError.description & @CRLF & _
            @TAB & "err.source is: " & @TAB & @TAB & $oError.source & @CRLF & _
            @TAB & "err.helpfile is: " & @TAB & $oError.helpfile & @CRLF & _
            @TAB & "err.helpcontext is: " & @TAB & $oError.helpcontext & @CRLF & _
            @TAB & "err.lastdllerror is: " & @TAB & $oError.lastdllerror & @CRLF & _
            @TAB & "err.scriptline is: " & @TAB & $oError.scriptline & @CRLF & _
            @TAB & "err.retcode is: " & @TAB & "0x" & Hex($oError.retcode) & @CRLF & @CRLF)
EndFunc   ;==>_ErrFunc
